@@ -38,6 +38,8 @@ import pickle
 import sys
 import unittest2 as unittest
 
+import mock
+
 from googleapiclient.discovery import _fix_up_media_upload
 from googleapiclient.discovery import _fix_up_method_description
 from googleapiclient.discovery import _fix_up_parameters
@@ -50,11 +52,14 @@ from googleapiclient.discovery import MEDIA_BODY_PARAMETER_DEFAULT_VALUE
 from googleapiclient.discovery import ResourceMethodParameters
 from googleapiclient.discovery import STACK_QUERY_PARAMETERS
 from googleapiclient.discovery import STACK_QUERY_PARAMETER_DEFAULT_VALUE
+from googleapiclient.discovery_cache import DISCOVERY_DOC_MAX_AGE
+from googleapiclient.discovery_cache.base import Cache
 from googleapiclient.errors import HttpError
 from googleapiclient.errors import InvalidJsonError
 from googleapiclient.errors import MediaUploadSizeError
 from googleapiclient.errors import ResumableUploadError
 from googleapiclient.errors import UnacceptableMimeTypeError
+from googleapiclient.errors import UnknownApiNameOrVersion
 from googleapiclient.http import BatchHttpRequest
 from googleapiclient.http import HttpMock
 from googleapiclient.http import HttpMockSequence
@@ -338,11 +343,18 @@ class DiscoveryErrors(unittest.TestCase):
   def test_failed_to_parse_discovery_json(self):
     self.http = HttpMock(datafile('malformed.json'), {'status': '200'})
     try:
-      plus = build('plus', 'v1', http=self.http)
+      plus = build('plus', 'v1', http=self.http, cache_discovery=False)
       self.fail("should have raised an exception over malformed JSON.")
     except InvalidJsonError:
       pass
 
+  def test_unknown_api_name_or_version(self):
+      http = HttpMockSequence([
+        ({'status': '404'}, open(datafile('zoo.json'), 'rb').read()),
+      ])
+      with self.assertRaises(UnknownApiNameOrVersion):
+        plus = build('plus', 'v1', http=http, cache_discovery=False)
+
 
 class DiscoveryFromDocument(unittest.TestCase):
 
@@ -366,6 +378,17 @@ class DiscoveryFromDocument(unittest.TestCase):
     plus = build_from_document(discovery, base=base)
     self.assertEquals("https://www.googleapis.com/plus/v1/", plus._baseUrl)
 
+  def test_building_with_optional_http(self):
+    discovery = open(datafile('plus.json')).read()
+    plus = build_from_document(discovery, base="https://www.googleapis.com/")
+    self.assertTrue(isinstance(plus._http, httplib2.Http))
+
+  def test_building_with_explicit_http(self):
+    http = HttpMock()
+    discovery = open(datafile('plus.json')).read()
+    plus = build_from_document(
+      discovery, base="https://www.googleapis.com/", http=http)
+    self.assertEquals(plus._http, http)
 
 class DiscoveryFromHttp(unittest.TestCase):
   def setUp(self):
@@ -402,6 +425,104 @@ class DiscoveryFromHttp(unittest.TestCase):
       self.assertEqual(e.uri, 'http://example.com')
 
 
+class DiscoveryFromAppEngineCache(unittest.TestCase):
+  def test_appengine_memcache(self):
+    # Hack module import
+    self.orig_import = __import__
+    self.mocked_api = mock.MagicMock()
+
+    def import_mock(name, *args):
+      if name == 'google.appengine.api':
+        return self.mocked_api
+      return self.orig_import(name, *args)
+
+    import_fullname = '__builtin__.__import__'
+    if sys.version_info[0] >= 3:
+      import_fullname = 'builtins.__import__'
+
+    with mock.patch(import_fullname, side_effect=import_mock):
+      namespace = 'google-api-client'
+      self.http = HttpMock(datafile('plus.json'), {'status': '200'})
+
+      self.mocked_api.memcache.get.return_value = None
+
+      plus = build('plus', 'v1', http=self.http)
+
+      # memcache.get is called once
+      url = 'https://www.googleapis.com/discovery/v1/apis/plus/v1/rest'
+      self.mocked_api.memcache.get.assert_called_once_with(url,
+                                                           namespace=namespace)
+
+      # memcache.set is called once
+      with open(datafile('plus.json')) as f:
+        content = f.read()
+      self.mocked_api.memcache.set.assert_called_once_with(
+        url, content, time=DISCOVERY_DOC_MAX_AGE, namespace=namespace)
+
+      # Returns the cached content this time.
+      self.mocked_api.memcache.get.return_value = content
+
+      # Make sure the contents are returned from the cache.
+      # (Otherwise it should through an error)
+      self.http = HttpMock(None, {'status': '200'})
+
+      plus = build('plus', 'v1', http=self.http)
+
+      # memcache.get is called twice
+      self.mocked_api.memcache.get.assert_has_calls(
+        [mock.call(url, namespace=namespace),
+         mock.call(url, namespace=namespace)])
+
+      # memcahce.set is called just once
+      self.mocked_api.memcache.set.assert_called_once_with(
+        url, content, time=DISCOVERY_DOC_MAX_AGE,namespace=namespace)
+
+
+class DictCache(Cache):
+  def __init__(self):
+    self.d = {}
+  def get(self, url):
+    return self.d.get(url, None)
+  def set(self, url, content):
+    self.d[url] = content
+  def contains(self, url):
+    return url in self.d
+
+
+class DiscoveryFromFileCache(unittest.TestCase):
+  def test_file_based_cache(self):
+    cache = mock.Mock(wraps=DictCache())
+    with mock.patch('googleapiclient.discovery_cache.file_cache.cache',
+                    new=cache):
+      self.http = HttpMock(datafile('plus.json'), {'status': '200'})
+
+      plus = build('plus', 'v1', http=self.http)
+
+      # cache.get is called once
+      url = 'https://www.googleapis.com/discovery/v1/apis/plus/v1/rest'
+      cache.get.assert_called_once_with(url)
+
+      # cache.set is called once
+      with open(datafile('plus.json')) as f:
+        content = f.read()
+      cache.set.assert_called_once_with(url, content)
+
+      # Make sure there is a cache entry for the plus v1 discovery doc.
+      self.assertTrue(cache.contains(url))
+
+      # Make sure the contents are returned from the cache.
+      # (Otherwise it should through an error)
+      self.http = HttpMock(None, {'status': '200'})
+
+      plus = build('plus', 'v1', http=self.http)
+
+      # cache.get is called twice
+      cache.get.assert_has_calls([mock.call(url), mock.call(url)])
+
+      # cahce.set is called just once
+      cache.set.assert_called_once_with(url, content)
+
+
 class Discovery(unittest.TestCase):
 
   def test_method_error_checking(self):
@@ -537,7 +658,7 @@ class Discovery(unittest.TestCase):
       ({'status': '200'}, 'echo_request_headers_as_json'),
       ])
     http = tunnel_patch(http)
-    zoo = build('zoo', 'v1', http=http)
+    zoo = build('zoo', 'v1', http=http, cache_discovery=False)
     resp = zoo.animals().patch(
         name='lion', body='{"description": "foo"}').execute()
 
@@ -1172,6 +1293,14 @@ class Next(unittest.TestCase):
     request = tasks.tasklists().list()
     self.assertEqual(None, tasks.tasklists().list_next(request, {}))
 
+  def test_next_successful_none_on_empty_page_token(self):
+    self.http = HttpMock(datafile('tasks.json'), {'status': '200'})
+    tasks = build('tasks', 'v1', http=self.http)
+    request = tasks.tasklists().list()
+    next_request = tasks.tasklists().list_next(
+        request, {'nextPageToken': ''})
+    self.assertEqual(None, next_request)
+
   def test_next_successful_with_next_page_token(self):
     self.http = HttpMock(datafile('tasks.json'), {'status': '200'})
     tasks = build('tasks', 'v1', http=self.http)
